unit UnitGame;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms,
  Dialogs, GLMisc, GLWin32Viewer, GLFireFX, GLObjects, GLScene, GLParticleFX,
  GLTexture, GLCadencer, GLMovement, GLExplosionFx, GLThorFx, GLSpaceText,
  VectorGeometry, Keyboard, Math, GLGeomObjects;

type
  TTargetType = (ttTrackTarget,ttActiveTarget);

  TFormGame = class(TForm)
    GLCadencer: TGLCadencer;
    GLScene: TGLScene;
    GLParticleRenderer: TGLParticleFXRenderer;
    GLCameraPlayer: TGLCamera;
    DummyCubePlayer: TGLDummyCube;
    GLFireManager: TGLFireFXManager;
    GLViewer: TGLSceneViewer;
    DummyCubeLevel: TGLDummyCube;
    GLLightSource: TGLLightSource;
    ArrowLinePlayer: TGLArrowLine;
    DummyPlayerFire: TGLDummyCube;
    ArrowLinePlayerUp: TGLArrowLine;
    procedure GLCadencerProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure DummyCubeLevelProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure DummyCubePlayerProgress(Sender: TObject; const deltaTime,
      newTime: Double);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
  public
    { Public declarations }
    Velocity: Single;
    MassFactor: Single;
    TargetInterval: Integer;
    SectorInterval: Single;
    SectorNumber: Integer;
    MaxOrizontalAngle,
    MaxVerticalAngle: Integer;

    RollStartLerpVector: TVector;
    RollEndLerpVector: TVector;
    LevelStartTime: Double;
    SectorStartTime: Double;

    TargetList: TList;
    HittedTarget: TList;
    Movement: TGLMovement;

    PlayerPosition: TVector;
    Procedure CreateLevel;
    Procedure DestroyLevel;
    function GenerateNode(Const ATrackPath: TGLMovementPath;
                          const AUp: TVector;
                          Const AVelocity: Single;
                          Const ATime: Double;
                          Const AGradAngleVertical: Integer = 0;
                          Const AGradAngleOrizontal: Integer = 0): TGLPathNode;
    function GenerateTarget(Const ALevelSceneObject: TGLBaseSceneObject;
                            Const ATrackPath: TGLMovementPath;
                            Const ATargetType: TTargetType;
                            Const AUp: TVector;
                            Const AColor: TGLColor;
                            Const AVelocity: Single;
                            Const ATime: Double;
                            Const AGradAngleVertical: Integer = 0;
                            Const AGradAngleOrizontal: Integer = 0;
                            Const APFXRederer: TGLParticleFXRenderer = nil;
                            Const AGradAngle: Integer = 0;
                            Const ARadialDistance: Single = 0;
                            Const ADimentionFactor: Single = 1): TGLDummyCube;
    Procedure DestroyTarget(Const ASceneObjectTarget: TGLBaseSceneObject);
    Procedure TargetProgress(Sender: TObject; const deltaTime, newTime: Double);
    Procedure DeleteTarget(Const ANodeID: Integer);
    Function FindTargetByNodeID(Const ANodeID: Integer): TGLBaseSceneObject;
    Procedure PathTravelStop(Sender: TObject);
    Procedure EnableTargetEffets(Const ATargetObject: TGLBaseSceneObject);
    Procedure DisableTargetEffets(Const ATargetObject: TGLBaseSceneObject);
  end;

var
  FormGame: TFormGame;

implementation

{$R *.dfm}

{ TFormGame }

procedure TFormGame.CreateLevel;
Var
  I: Integer;
  AColor: TGLColor;
  PrevTargetObject,
  TargetObject: TGLBaseSceneObject;
begin
  LevelStartTime := -1;

  Velocity := 6;
  MassFactor := 1;
  TargetInterval := 3;
  SectorInterval := 5;
  SectorNumber := 50;
  MaxOrizontalAngle := 45;
  MaxVerticalAngle := 45;

  TargetList := TList.Create;
  HittedTarget := TList.Create;

  Movement := GetOrCreateMovement(DummyCubePlayer);
  Movement.OnAllPathTravelledOver := PathTravelStop;
  With Movement.AddPath Do
  Begin
    //ShowPath := True;
    Movement.ActivePathIndex := Index;
    With AddNodeFromObject(DummyCubeLevel) Do
      Speed := Velocity;
    With AddNodeFromObject(DummyCubePlayer) Do
      Speed := Velocity;
  End;
  GenerateTarget(DummyCubeLevel,
                Movement.ActivePath,
                ttTrackTarget,
                DummyCubeLevel.Up.AsVector,
                ArrowLinePlayer.Material.FrontProperties.Diffuse
                Velocity,
                SectorInterval,
                0,
                0,
                GLParticleRenderer);
  GenerateTarget(DummyCubeLevel,
                Movement.ActivePath,
                ttTrackTarget,
                DummyCubeLevel.Up.AsVector,
                ArrowLinePlayer.Material.FrontProperties.Diffuse
                Velocity,
                SectorInterval,
                0,
                0,
                GLParticleRenderer);

  For I := 0 To SectorNumber-1 Do
  Begin
    AColor := TGLColor.Create(Nil);
    Try
      AColor.Alpha := 1;
      AColor.Blue := Random(10)/10;
      AColor.Red := Random(10)/10;
      AColor.Green := Random(10)/10;
      If (I mod TargetInterval) = 0 Then
        TargetObject :=  GenerateTarget(DummyCubeLevel,
                                        Movement.ActivePath,
                                        ttActiveTarget,
                                        DummyCubeLevel.Up.AsVector,
                                        AColor,
                                        Velocity,
                                        SectorInterval,
                                        RandomRange(-MaxVerticalAngle,MaxVerticalAngle),
                                        RandomRange(-MaxOrizontalAngle,MaxOrizontalAngle),
                                        GLParticleRenderer,
                                        RandomRange(0,360),
                                        0.8,
                                        0.2)
      Else
        TargetObject :=  GenerateTarget(DummyCubeLevel,
                                        Movement.ActivePath,
                                        ttTrackTarget,
                                        DummyCubeLevel.Up.AsVector,
                                        AColor,
                                        Velocity,
                                        SectorInterval,
                                        RandomRange(-MaxVerticalAngle,MaxVerticalAngle),
                                        RandomRange(-MaxOrizontalAngle,MaxOrizontalAngle),
                                        GLParticleRenderer,
                                        0,
                                        1,
                                        1);
      PrevTargetObject := FindTargetByNodeID(Movement.ActivePath.Nodes.Items[Movement.ActivePath.NodeCount-2].ID);
      If PrevTargetObject<>nil Then
        PrevTargetObject.Direction := TargetObject.Direction;
    Finally
      AColor.Free;
    End;
    TargetList.Add(TargetObject);
    TargetObject.Tag := Movement.ActivePath.Nodes[Movement.ActivePath.NodeCount-1].ID;
  End;
end;

procedure TFormGame.DeleteTarget(const ANodeID: Integer);
Var TargetObject: TGLBaseSceneObject;
begin
  TargetObject := FindTargetByNodeID(ANodeID);
  If TargetObject<>nil Then
  Begin
    DestroyTarget(TargetObject);
    TargetList.Remove(TargetObject);
  End;
end;

procedure TFormGame.DestroyLevel;
begin
  While TargetList.Count>0 Do
    DeleteTarget(TGLBaseSceneObject(TargetList[0]).Tag);
  Movement.StopPathTravel;
  Movement.ClearPaths;
  HittedTarget.Free;
  TargetList.Free;
end;

function TFormGame.GenerateNode(const ATrackPath: TGLMovementPath; const AUp: TVector;
  const AVelocity: Single; const ATime: Double; const AGradAngleVertical,
  AGradAngleOrizontal: Integer):  TGLPathNode;
Var
  ModuleFactor: Single;
  AngleFactor: Single;
  VectorPosition,
  VectorRotation,
  Vector: TVector;
begin
  ModuleFactor := ATime;
  AngleFactor := (Pi/180);

  VectorPosition := VectorSubtract(ATrackPath.Nodes.Items[ATrackPath.Nodes.Count-1].PositionAsVector,
                                   ATrackPath.Nodes.Items[ATrackPath.Nodes.Count-2].PositionAsVector);

  If AGradAngleVertical<>0 Then
  Begin
    RotateVector(VectorPosition,AUp,AGradAngleVertical*AngleFactor);
  End;
  If  AGradAngleOrizontal<>0 Then
  Begin
    VectorCrossProduct(AUp,VectorPosition,Vector);
    RotateVector(VectorPosition,Vector,AGradAngleOrizontal*AngleFactor);
  End;
  NormalizeVector(VectorPosition);
  Result := ATrackPath.AddNode;
  With Result Do
  Begin
    VectorPosition := VectorScale(VectorPosition,ModuleFactor);
    VectorPosition := VectorAdd(ATrackPath.Nodes.Items[ATrackPath.Nodes.Count-1].PositionAsVector,
                                VectorPosition);
    PositionAsVector := VectorPosition;
    Speed := AVelocity;
  End;
end;

function TFormGame.GenerateTarget(Const ALevelSceneObject: TGLBaseSceneObject;
                                  Const ATrackPath: TGLMovementPath;
                                  Const ATargetType: TTargetType;
                                  Const AUp: TVector;
                                  Const AColor: TGLColor;
                                  Const AVelocity: Single;
                                  Const ATime: Double;
                                  Const AGradAngleVertical: Integer = 0;
                                  Const AGradAngleOrizontal: Integer = 0;
                                  Const APFXRederer: TGLParticleFXRenderer = nil;
                                  Const AGradAngle: Integer = 0;
                                  Const ARadialDistance: Single = 0;
                                  Const ADimentionFactor: Single = 1): TGLDummyCube;
Var
  PathNode: TGLPathNode;
  SceneObject: TGLBaseSceneObject;
  VectorDirection,
  Vector1,
  Vector2: TVector;
  EffectMenager: TGLPolygonPFXManager;
  I: Integer;
begin
    PathNode := GenerateNode(ATrackPath,
                              AUp,
                              AVelocity,
                              ATime,
                              AGradAngleVertical,
                              AGradAngleOrizontal);
    VectorDirection := VectorSubtract(ATrackPath.Nodes.Items[ATrackPath.Nodes.Count-1].PositionAsVector,
                                     ATrackPath.Nodes.Items[ATrackPath.Nodes.Count-2].PositionAsVector);
    NormalizeVector(VectorDirection);
    Result := TGLDummyCube(ALevelSceneObject.AddNewChildFirst(TGLDummyCube));
    Result.Position.AsVector := PathNode.PositionAsVector;
    Result.Direction.AsVector := VectorDirection;
    PathNode.RotationAsVector := Result.Rotation.AsVector;
    Case ATargetType Of
      ttTrackTarget:
      Begin
        SceneObject := Result.AddNewChildFirst(TGLTorus);
        TGLTorus(SceneObject).MajorRadius := ARadialDistance;
        TGLTorus(SceneObject).MinorRadius := (1/15)*TGLTorus(SceneObject).MajorRadius;
        PathNode.RotationAsVector := SceneObject.Rotation.AsVector;
        SceneObject.Scale.AsVector := VectorScale(SceneObject.Scale.AsVector,ADimentionFactor);
        TGLSceneObject(SceneObject).Material.FrontProperties.Emission := AColor;
      End;
      ttActiveTarget:
      Begin
        SceneObject := Result.AddNewChildFirst(TGLDummyCube);
        Vector1 := VectorCrossProduct(AUp,SceneObject.Direction.AsVector);
        Vector2 := VectorCrossProduct(Vector1,SceneObject.Direction.AsVector);
        NormalizeVector(Vector1);
        NormalizeVector(Vector2);
        RotateVector(Vector1,VectorDirection,RadToGrad(AGradAngle));
        Vector1 := VectorScale(Vector1,ARadialDistance);
        SceneObject.Position.AsVector := Vector1;
        SceneObject.Scale.AsVector := VectorScale(SceneObject.Scale.AsVector,ADimentionFactor);
        with GetOrCreateSourcePFX(SceneObject) do
        begin
           EffectMenager := TGLPolygonPFXManager.Create(nil);
           With EffectMenager Do
           Begin
             ParticleSize := 0.12;
             ColorInner := AColor;
             ColorOuter.Alpha := 0;
             With LifeColors.Add Do
             Begin
               ColorInner := AColor;
               ColorOuter.Alpha := 0;
               LifeTime := 0.5;
             End;
             //Renderer := APFXRederer;
           End;
           For I := 0 To APFXRederer.Scene.Owner.ComponentCount-1 Do
             If (APFXRederer.Scene.Owner.Components[I] Is TGLCadencer)
                 And (TGLCadencer(APFXRederer.Scene.Owner.Components[I]).Scene.Name = APFXRederer.Scene.Name) Then
               EffectMenager.Cadencer := TGLCadencer(APFXRederer.Scene.Owner.Components[I]);
           Manager := EffectMenager;
           VelocityDispersion:=0;
           ParticleInterval := 0;
        end;
        SceneObject := SceneObject.AddNewChildFirst(TGLFrustrum);
        TGLSceneObject(SceneObject).Material.FrontProperties.Emission := AColor;
        With TGLFrustrum(SceneObject) Do
        Begin
          Height := 1;
          ApexHeight := 1;
          BaseWidth := 1;
          BaseDepth := 1;
        End;
        SceneObject := SceneObject.AddNewChildFirst(TGLFrustrum);
        TGLSceneObject(SceneObject).Material.FrontProperties.Emission := AColor;
        With TGLFrustrum(SceneObject) Do
        Begin
          Height := 1;
          ApexHeight := 1;
          BaseWidth := 1;
          BaseDepth := 1;
          RollAngle := 180;
        End;
        Result.OnProgress := TargetProgress;
      End;
    End;
end;

procedure TFormGame.TargetProgress(Sender: TObject; const deltaTime,
  newTime: Double);
Var J: Integer;
begin
  TGLBaseSceneObject(Sender).Children[0].TurnAngle := 4*NewTime*60;
  If TGLBaseSceneObject(Sender).DistanceTo(DummyCubePlayer)<GLViewer.Buffer.FogEnvironment.FogEnd Then
    EnableTargetEffets(TGLBaseSceneObject(Sender))
  Else
    DisableTargetEffets(TGLBaseSceneObject(Sender));
end;

procedure TFormGame.GLCadencerProgress(Sender: TObject; const deltaTime,
  newTime: Double);
begin
  Randomize;
  GLViewer.Invalidate;
end;

procedure TFormGame.DummyCubeLevelProgress(Sender: TObject;
  const deltaTime, newTime: Double);
begin
  If LevelStartTime < 0 Then
    If IsKeyDown(VK_Space) Then
    Begin
      LevelStartTime := NewTime;
      SectorStartTime := NewTime;
      Movement.StartPathTravel;
    End Else
      Exit;
  If IsKeyDown(Char('h')) Then
    //

end;

procedure TFormGame.FormCreate(Sender: TObject);
begin
  CreateLevel;
end;

procedure TFormGame.FormDestroy(Sender: TObject);
begin
  GLCadencer.OnProgress := nil;
  GLCadencer.Enabled := False;
end;

procedure TFormGame.FormShow(Sender: TObject);
begin
  GLCadencer.Enabled := true;
end;

procedure TFormGame.DummyCubePlayerProgress(Sender: TObject;
  const deltaTime, newTime: Double);
begin
  DummyCubePlayer.Direction.AsVector := VectorSubtract(DummyCubePlayer.Position.AsVector,PlayerPosition);
  PlayerPosition := DummyCubePlayer.Position.AsVector;
end;

procedure TFormGame.PathTravelStop(Sender: TObject);
begin
  LevelStartTime := -1;
end;

procedure TFormGame.DestroyTarget(Const ASceneObjectTarget: TGLBaseSceneObject);
Var
  J: Integer;
begin
  If ASceneObjectTarget Is TGLBaseSceneObject Then
    With TGLBaseSceneObject(ASceneObjectTarget) Do
    Begin
      For J := Effects.Count-1 DownTo 0 Do
      Begin
        If Effects[J] Is TGLSourcePFXEffect Then
        Begin
          TGLSourcePFXEffect(Effects[J]).Manager.Cadencer := nil;
          TGLSourcePFXEffect(Effects[J]).Manager.Free;
        End;
        If Effects[J] Is TGLBFireFX Then
        Begin
          TGLBFireFX(Effects[J]).Manager.Cadencer := nil;
          TGLBFireFX(Effects[J]).Manager.Free;
        End;
        If Effects[J] Is TGLBThorFX Then
        Begin
          TGLBThorFX(Effects[J]).Manager.Cadencer := nil;
          TGLBThorFX(Effects[J]).Manager.Free;
        End;
        Effects.Delete(J);
      End;
    End;

  For J := 0 To ASceneObjectTarget.Count-1 Do
    If ASceneObjectTarget.Children[J].Count>0 Then
      DestroyTarget(ASceneObjectTarget.Children[J]);
end;

procedure TFormGame.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  DestroyLevel;
end;

Function TFormGame.FindTargetByNodeID(Const ANodeID: Integer): TGLBaseSceneObject;
Var I: Integer;
begin
  Result := nil;
  For I:= 0 To (TargetList.Count-1) Do
    If TGLBaseSceneObject(TargetList[I]).Tag = ANodeID Then
    Begin
      Result := TGLBaseSceneObject(TargetList[I]);
      Break;
    End;
end;

procedure TFormGame.EnableTargetEffets(
  const ATargetObject: TGLBaseSceneObject);
Var
  J: Integer;
begin
  With ATargetObject Do
    For J := Effects.Count-1 DownTo 0 Do
    Begin
      If Effects[J] Is TGLSourcePFXEffect Then
      Begin
        With TGLSourcePFXEffect(Effects[J]) Do
        Begin
          Manager.Renderer := GLParticleRenderer;
          VelocityDispersion:=0.20;
          ParticleInterval := 0.04;
        End;
      End;
      If Effects[J] Is TGLBFireFX Then
        TGLBFireFX(Effects[J]).Manager.Disabled := false;
      If Effects[J] Is TGLBThorFX Then
        TGLBThorFX(Effects[J]).Manager.Disabled := false;
    End;
  For J := 0 To ATargetObject.Count-1 Do
    If ATargetObject.Children[J].Count>0 Then
      EnableTargetEffets(ATargetObject.Children[J]);
end;

procedure TFormGame.DisableTargetEffets(
  const ATargetObject: TGLBaseSceneObject);
Var
  J: Integer;
begin
  With ATargetObject Do
    For J := Effects.Count-1 DownTo 0 Do
    Begin
      If Effects[J] Is TGLSourcePFXEffect Then
      Begin
        With TGLSourcePFXEffect(Effects[J]) Do
        Begin
          Manager.Renderer := nil;
          VelocityDispersion:=0;
          ParticleInterval := 0;
        End;
      End;
      If Effects[J] Is TGLBFireFX Then
        TGLBFireFX(Effects[J]).Manager.Disabled := true;
      If Effects[J] Is TGLBThorFX Then
        TGLBThorFX(Effects[J]).Manager.Disabled := true;
    End;
  For J := 0 To ATargetObject.Count-1 Do
    If ATargetObject.Children[J].Count>0 Then
      DisableTargetEffets(ATargetObject.Children[J]);
end;
end.
